using System;
using System.Collections;

/* Weight of FibonacciNodes should be generic.
 * Generics themselves are crappy for this Heap, because Copmarison 
 * operators are assured about the interface IComparable, but + operator
 * does'nt have an interface. Because of that a generic class had to
 * encapsulate a value type and additionally support a user defined
 * interface like "IAddable" or sth.
 * 
 * Instead, the following statement should be sufficient for our needs. */
using TWeight = System.Decimal;

public class Dijkstra
{
	private Dijkstra() {}

    public static int[] CreateRoute(TWeight[,] m, int start, int dest, int[] stopovers)
    {
        if (stopovers == null || stopovers.Length == 0)
            return CreateRoute(m, start, dest);

        /* split stopovers into single paths
         * handling start and destination extra */
        int[][] singlePaths = new int[stopovers.Length + 1][];

        singlePaths[0] = CreateRoute(m, start, stopovers[0]);

        for (int i=1; i < stopovers.Length; i++)
            singlePaths[i] = CreateRoute(m, stopovers[i - 1], stopovers[i]);
        
        singlePaths[stopovers.Length] = CreateRoute(m, stopovers[stopovers.Length - 1], dest);
        
        // Check, if any of the single paths is not valid (== one way / end of road)
		for (int i = 0; i < singlePaths.Length; i++)
			if (singlePaths[i] == null)
				return null;

        // Count nr of all nodes of all paths
        int nrOfNodes = 0;
        for (int i=0; i<singlePaths.Length; i++)
            nrOfNodes += singlePaths[i].Length;
		nrOfNodes -= stopovers.Length;

        // Copy nodes of all paths into resulting path
        int[] resultPath = new int[nrOfNodes];
		int posResult = 0;
        for (int i=0; i<singlePaths.Length; i++)
			for (int ii=0; ii<singlePaths[i].Length-1; ii++)
				resultPath[posResult++] = singlePaths[i][ii];
		resultPath[posResult] = dest;
        
        return resultPath;
    }

    public static int[] CreateRoute(TWeight[,] m, int start, int dest)
	{
        /* Create an array of nodes 
         * Each call of Heap.Insert() will fill beside the heap also
         * this array. It will grant access to the nodes in the middle of the tree. */
        FibonacciNode[] nodes = new FibonacciNode[m.GetLength(0)];

        /* 1. Preparations
         * Create Fibonacci Heap and fill it with nodes
         * Each reference is copied into the array of nodes 
         * Each node gets a key of -1 except nodes[0] */
        FibonacciHeap f = new FibonacciHeap();
		for (int i=0; i<m.GetLength(0); i++)
		{
			int key = (i == start) ? 0 : -1;
			nodes[i] = f.Insert(key, i, null);
		}

        FibonacciNode fn = null;
		while (!f.Empty())
        {
            /* 2. Extracting means no need for visited-flag
             *    Visited nodes are not any longer within the heap */
			fn = f.ExtractMinNode();

			if (fn.key < 0)
				return null;

            // (Wikipedia): The destination has been reached -> break
            if (fn.id == dest)
                break;
                
            // 4. Aquire all successors of current TrackItem
    		for (int i=0; i<m.GetLength(0); i++)
            {
				// Check, if connected
				if (m[fn.id, i] > 0)
				{
					/* 4. Set new weight if connected successor has no route predessor or
                     * 5. If successor's weight > node's weight + distance between node and successor, 
                     *    actual way along node is better */
					if (nodes[i].key == -1 || nodes[i].key > fn.key + m[fn.id, i])
					{
						// New key = key of current visited node + distance between node and successor, taken from matrix
						TWeight nuKey = fn.key + m[fn.id, i];
						f.ReplaceKey(nodes[i], nuKey);
                        nodes[i].predessor = fn;
    				}
				}
            }
        }

        // Create an empty path container
        ArrayList path = new ArrayList();

		// Build path from destination backwards by following every item's predessor
		if (fn == null)
			return null;

		while(fn.predessor != null && fn.id != start)
		{
			path.Add(fn);
            fn = fn.predessor;			
		}

		if (path.Count == 0)
			return null;
		else
		{
			/* Return the path as array of ints
             * This array must be filled reversed */
			int[] pathI = new int[path.Count+1];
			pathI[0] = start;
            for (int i=1, ii=path.Count-1; i<=path.Count; i++, ii--)
				pathI[i] = ((FibonacciNode) path[ii]).id;
			return pathI;
		}
	}
}