using System;

/* Weight of FibonacciNodes should be generic.
 * Generics themselves are crappy for this Heap, because Comparison 
 * operators are assured about the interface IComparable, but + operator
 * doesn't have an interface. Because of that a generic class had to
 * encapsulate a value type and additionally support a user defined
 * interface like "IAddable" or sth equal.
 * 
 * Instead, the following statement should be sufficient for our needs. */
using TWeight = System.Decimal;

class FibonacciHeap
{
	// Members
	private int nrOfNodes;
	private FibonacciNode min;
	public FibonacciNode root;

	// C'tor
	public FibonacciHeap() {}

	// Methods
	public int Compare(FibonacciNode a, FibonacciNode b)
	{
		// -1 means infinity; this must be checked before usual compare
		if (a.key == -1 && b.key != -1)
			return 1;
		if (a.key != -1 && b.key == -1)
			return -1;
		if (a.key == b.key)
			return 0;

		// usual compare
		if (a.key < b.key)
			return -1;
		if (a.key > b.key)
			return 1;
		return 0;
	}

	public int CompareData(TWeight key, Object data, FibonacciNode b)
	{
		FibonacciNode a = new FibonacciNode();
		a.key = key;
		a.data = data;
		return this.Compare(a, b);
	}

	/* Insert DATA, with priority KEY, into HEAP.  */
	public FibonacciNode Insert(TWeight key, int id, Object data)
	{
		FibonacciNode node = new FibonacciNode(key, id, data);
		this.InsertRoot(node);

		/* If their was no minimum, or this key is less than the min, it's the new min.  */
		//if (heap.min == null || node.key < heap.min.key)	// old comparison.
		// new: infinity == -1
		if (this.min == null || this.min.key == -1 || (node.key != -1 && node.key < this.min.key))
			this.min = node;

		this.nrOfNodes++;

		return node;		
	}

	/* Return the data of the minimum node (if we know it).  */
	public Object MinData()
	{
		/* If there is no min, we can't easily return it.  */
		if (this.min == null)
			return null;
		return this.min.data;
	}

	/* Return the key of the minimum node (if we know it).  */
	public TWeight MinKey()
	{
		/* If there is no min, we can't easily return it.  */
		if (this.min == null)
			return 0;
		return this.min.key;
	}

	/* Union HEAPA and HEAPB into a new heap.  */
	/* not implemented */

	/* Extract the data of the minimum node from HEAP.  */
	public Object ExtractMinData ()
	{
		FibonacciNode z;
		Object ret = null;

		/* If we don't have a min set, it means we have no nodes.  */
		if (this.min != null)
		{
			/* Otherwise, extract the min node, free the node, and return the node's data.  */
			z = ExtractMinNode();
			ret = z.data;
			// freeing of the node should do the .net garbage collector
		}

		return ret;
	}

	/* Replace both the KEY and the DATA associated with NODE.  */
	public Object ReplaceKeyData(FibonacciNode node, TWeight key, Object data)
	{
		Object odata;
		TWeight okey;
		FibonacciNode y;

		/* If we wanted to, we could actually do a real increase by redeleting and
		inserting. However, this would require O (log n) time. So just bail out
		for now.  */
		if (CompareData(key, data, node) > 0)
			return null;

		odata = node.data;
		okey = node.key;
		node.data = data;
		node.key = key;
		y = node.parent;

		if (okey == key)
			return odata;

		/* These two compares are specifically <= 0 to make sure that in the case
		of equality, a node we replaced the data on, becomes the new min.  This
		is needed so that delete's call to extractmin gets the right node.  */
		if (y != null && Compare(node, y) <= 0)
		{
			this.Cut(node, y);
			this.CascadingCut(y);
		}

		if (Compare(node, this.min) <= 0)
			this.min = node;

		return odata;
	}

	/* Replace the DATA associated with NODE.  */
	public Object ReplaceData(FibonacciNode node, Object data)
	{
		return ReplaceKeyData(node, node.key, data);
	}

	/* Replace the KEY associated with NODE.  */
    public TWeight ReplaceKey(FibonacciNode node, TWeight key)
	{
		TWeight okey = node.key;
		ReplaceKeyData(node, key, node.data);
		return okey;
	}

	/* Delete NODE from HEAP.  */
	public Object DeleteNode(FibonacciNode node)
	{
		Object ret = node.data;

		/* To perform delete, we just make it the min key, and extract.  */
		ReplaceKey(node, -2);
		ExtractMinData ();

		return ret;
	}

	/* Delete HEAP.  */
	// Necessary in .net? or does this the garbage collector? I think yes.
	// Therefore not implemented

	/* Determine if HEAP is empty.  */
	public bool Empty()
	{
		return this.nrOfNodes == 0;
	}

	/* Extract the minimum node of the heap.  */
	public FibonacciNode ExtractMinNode()
	{
		FibonacciNode ret = this.min;
		FibonacciNode x, y, orig;

		/* Attach the child list of the minimum node to the root list of the heap.
		If there is no child list, we don't do squat.  */
		for (x = ret.child, orig = null; x != orig && x != null; x = y)
		{
			if (orig == null)
			orig = x;
			y = x.right;
			x.parent = null;
			this.InsertRoot(x);
		}

		/* Remove the old root.  */
		this.RemoveRoot(ret);
		this.nrOfNodes--;

		/* If we are left with no nodes, then the min is NULL.  */
		if (this.nrOfNodes == 0)
			this.min = null;
		else
		{
			/* Otherwise, consolidate to find new minimum, as well as do the reorg
				work that needs to be done.  */
			this.min = ret.right;
			this.Consolidate();
		}

		return ret;
	}

	/* Insert NODE into the root list of HEAP.  */
	private void InsertRoot(FibonacciNode node)
	{
		/* If the heap is currently empty, the new node becomes the singleton
		circular root list.  */
		if (this.root == null)
		{
            this.root = node;
			node.left = node;
			node.right = node;
			return;
		}

		/* Otherwise, insert it in the circular root list between the root
			and it's right node.  */
		this.InsertAfter(this.root, node);
	}

	/* Remove NODE from the rootlist of HEAP.  */
	private void RemoveRoot(FibonacciNode node)
	{
		if (node.left == node)
			this.root = null;
		else
			this.root = this.Remove(node);
	}

	/* Consolidate the heap.  */
	private void Consolidate()
	{
		int i;
		int d;
		int D;
		
		//FibonacciNode[] a = new FibonacciNode[1 + 8 * sizeof (long)];
		FibonacciNode[] a = new FibonacciNode[1 + 8 * 4];
		FibonacciNode w = new FibonacciNode();
		FibonacciNode y = new FibonacciNode();
		FibonacciNode x = new FibonacciNode();
		

		D = 1 + 8 * 4;// * sizeof (long);

		// memset (a, 0, sizeof (fibnode_t) * D);

		while ((w = this.root) != null)
		{
			x = w;
			this.RemoveRoot(w);
			d = x.degree;
			while (a[d] != null)
			{
				y = a[d];
				if (Compare(x, y) > 0)
				{
					FibonacciNode temp;
					temp = x;
					x = y;
					y = temp;
				}
				this.Link(y, x);
				a[d] = null;
				d++;
			}
		a[d] = x;
		}
		this.min = null;
		for (i = 0; i < D; i++)
			if (a[i] != null)
			{
				this.InsertRoot(a[i]);
				if (this.min == null || Compare(a[i], this.min) < 0)
				this.min = a[i];
			}
	}

	/* Make NODE a child of PARENT.  */
	private void Link(FibonacciNode node, FibonacciNode parent)
	{
		if (parent.child == null)
			parent.child = node;
		else
			this.InsertBefore(parent.child, node);
		node.parent = parent;
		parent.degree++;
		node.mark = 0;
	}

	/* Remove NODE from PARENT's child list.  */
	private void Cut(FibonacciNode node, FibonacciNode parent)
	{
		this.Remove(node);
		parent.degree--;
		this.InsertRoot(node);
		node.parent = null;
		node.mark = 0;
	}

	private void CascadingCut(FibonacciNode y)
	{
		FibonacciNode z;

		while ((z = y.parent) != null)
		{
			if (y.mark == 0)
			{
				y.mark = 1;
				return;
			}
			else
			{
				this.Cut(y, z);
				y = z;
			}
		}
	}

	private void InsertAfter(FibonacciNode a, FibonacciNode b)
	{
		if (a == a.right)
		{
			a.right = b;
			a.left = b;
			b.right = a;
			b.left = a;
		}
		else
		{
			b.right = a.right;
			a.right.left = b;
			a.right = b;
			b.left = a;
		}
	}

	public FibonacciNode Remove(FibonacciNode node)
	{
		FibonacciNode ret;

		if (node == node.left)
			ret = null;
		else
			ret = node.left;

		if (node.parent != null && node.parent.child == node)
			node.parent.child = ret;

		node.right.left = node.left;
		node.left.right = node.right;

		node.parent = null;
		node.left = node;
		node.right = node;

		return ret;
	}

	private void InsertBefore(FibonacciNode a, FibonacciNode b)
	{
		// #define fibnode_insert_before(a, b) fibnode_insert_after (a->left, b)
		this.InsertAfter(a.left, b);
	}
}

class FibonacciNode
{
	// Members
	public FibonacciNode parent;    // The parent related to the tree / heap
	public FibonacciNode child;
	public FibonacciNode left;
	public FibonacciNode right;
	public FibonacciNode predessor; // The predessor related to the path (Dijkstra)
	public TWeight key;
	public int id;
	public Object data;             // Not used yet

	public int degree = 0;	// : 31;
	public int mark = 0;	// : 1;

    /* Properties
     * The FibonacciNode has nothing to hide. 
     * So the use of properties is not necessary. */

	// C'tor
	public FibonacciNode()
	{
		this.left = this.right = this;
        this.key = 0;
        this.id = 0;
        this.data = null;
	}

	public FibonacciNode(TWeight key, Object data)
	{
		this.left = this.right = this;
		this.key = key;
        this.id = 0;
		this.data = data;
	}

	public FibonacciNode(TWeight key, int id, Object data)
	{
		this.left = this.right = this;
		this.key = key;
		this.id = id;
		this.data = data;
	}
}